package net.knarcraft.blockhunt.listener;

import me.libraryaddict.disguise.DisguiseAPI;
import net.knarcraft.blockhunt.BlockHunt;
import net.knarcraft.blockhunt.MemoryStorage;
import net.knarcraft.blockhunt.arena.Arena;
import net.knarcraft.blockhunt.arena.ArenaState;
import net.knarcraft.blockhunt.config.MessageKey;
import net.knarcraft.blockhunt.util.ArenaHelper;
import net.knarcraft.blockhunt.util.MessageHelper;
import net.knarcraft.blockhunt.util.PlayerHelper;
import net.knarcraft.blockhunt.util.StringHelper;
import net.knarcraft.blockhunt.util.TokensHelper;
import net.knarcraft.blockhunt.util.VisibilityHelper;
import org.bukkit.GameMode;
import org.bukkit.Sound;
import org.bukkit.entity.Arrow;
import org.bukkit.entity.Player;
import org.bukkit.event.EventHandler;
import org.bukkit.event.EventPriority;
import org.bukkit.event.Listener;
import org.bukkit.event.entity.EntityDamageByEntityEvent;
import org.bukkit.event.entity.EntityDamageEvent.DamageCause;
import org.bukkit.projectiles.ProjectileSource;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

public class OnEntityDamageByEntityEvent implements Listener {

    @EventHandler(priority = EventPriority.NORMAL, ignoreCancelled = true)
    public void onEntityDamageByEntityEvent(EntityDamageByEntityEvent event) {
        if (!(event.getEntity() instanceof Player player)) {
            // We only care about player damage
            return;
        }

        // Early exit if no one is in any arena
        if (ArenaHelper.noPlayersInArenas()) {
            return;
        }

        Player damageDealer = null;
        if (event.getDamager() instanceof Player) {
            damageDealer = (Player) event.getDamager();
        } else {
            if ((event.getCause() == DamageCause.PROJECTILE) && (event.getDamager() instanceof Arrow arrow)) {
                // If damage was caused by an arrow, find out who shot the arrow
                ProjectileSource shooter = arrow.getShooter();
                if (shooter instanceof Player) {
                    damageDealer = (Player) shooter;
                }
            }
        }

        // Always block all damage not dealt by a player
        if (damageDealer == null) {
            event.setCancelled(true);
            return;
        }

        Arena arena = MemoryStorage.playerArena.get(player);
        if (arena == null) {
            return;
        }

        if (arena.gameState == ArenaState.WAITING || arena.gameState == ArenaState.STARTING) {
            // Always cancel damage when players are waiting
            event.setCancelled(true);
            return;
        }

        // Seeker receiving damage
        if (arena.seekers.contains(player)) {
            if (arena.seekers.contains(damageDealer)) {
                // Seeker damaged by seeker
                if (!arena.seekersCanHurtSeekers) {
                    event.setCancelled(true);
                    return;
                }
            } else {
                // Seeker damaged by hider
                if (!arena.hidersCanHurtSeekers) {
                    event.setCancelled(true);
                    return;
                }
            }
        } else {
            // Hider damaged by hider
            if (!arena.hidersCanHurtHiders && !arena.seekers.contains(damageDealer)) {
                event.setCancelled(true);
                return;
            }
        }

        // The damage is allowed, so lets handle it!
        player.getWorld().playSound(player.getLocation(), Sound.ENTITY_PLAYER_HURT, 1, 1);

        if (event.getDamage() < player.getHealth()) {
            return;
        }

        player.setHealth(20);
        event.setCancelled(true);

        // try moving everything to the next tick to prevent "Removing entity while ticking" errors
        Player finalDamageDealer = damageDealer;
        player.getServer().getScheduler().runTask(BlockHunt.plugin, () -> {
            killPlayer(arena, player, finalDamageDealer);
        });
    }

    /**
     * "Kills" a player in an arena
     *
     * @param arena             <p>The arena the player is in</p>
     * @param player            <p>The player to "kill"</p>
     * @param finalDamageDealer <p>The player that caused the last bit of damage</p>
     */
    private void killPlayer(Arena arena, Player player, Player finalDamageDealer) {
        DisguiseAPI.undisguiseToAll(player);
        MemoryStorage.playerBlock.remove(player);

        if (arena.hiders.contains(player)) {
            killHider(arena, player, finalDamageDealer);
        }

        player.getInventory().clear();

        if (arena.seekers.size() >= arena.playersInArena.size()) {
            ArenaHelper.seekersWin(arena);
        } else {
            DisguiseAPI.undisguiseToAll(player);
            MemoryStorage.seekerTime.put(player, arena.waitingTimeSeeker);
            PlayerHelper.teleport(player, arena.arenaLocations.getSeekersWarp());
            player.setGameMode(GameMode.SURVIVAL);
            player.setWalkSpeed(0.3F);

            // Fix for client not showing players after they join
            VisibilityHelper.makePlayerVisible(arena, player);
        }
    }

    /**
     * "Kills" a hider in an arena
     *
     * @param arena             <p>The arena the player is in</p>
     * @param player            <p>The hider to "kill"</p>
     * @param finalDamageDealer <p>The player that caused the last bit of damage</p>
     */
    private void killHider(Arena arena, Player player, Player finalDamageDealer) {
        int damageDealerTokens = TokensHelper.getTokens(finalDamageDealer);
        TokensHelper.setTokens(finalDamageDealer, damageDealerTokens + arena.killTokens);

        MessageHelper.sendMessage(finalDamageDealer, MessageKey.NORMAL_ADDED_TOKEN, "amount-" + arena.killTokens);

        int playerTokens = TokensHelper.getTokens(player);
        float addedTokens = ((float) arena.hidersTokenWin - (((float) arena.timer / (float) arena.gameTime) *
                (float) arena.hidersTokenWin));
        TokensHelper.setTokens(player, playerTokens + (int) addedTokens);

        MessageHelper.sendMessage(player, MessageKey.NORMAL_ADDED_TOKEN, "amount-" + (int) addedTokens);

        arena.seekers.add(player);
        player.setWalkSpeed(0.3F);
        ArenaHelper.sendMessage(arena, MessageKey.NORMAL_IN_GAME_HIDER_DIED,
                "%playerName%-" + player.getDisplayName(), "killer-" + finalDamageDealer.getDisplayName());

        int hiderCount = arena.hiders.size();
        if ((hiderCount <= 3) && (hiderCount > 0)) {
            List<String> hiders = new ArrayList<>();
            for (Player p : arena.playersInArena) {
                if (!arena.seekers.contains(p)) {
                    hiders.add(p.getName());
                }
            }
            Collections.sort(hiders);

            ArenaHelper.sendMessage(arena, MessageKey.NORMAL_IN_GAME_HIDERS_LEFT,
                    "left-" + String.join(", ",
                            StringHelper.toString(hiders.toArray())));
        } else {
            ArenaHelper.sendMessage(arena, MessageKey.NORMAL_IN_GAME_HIDERS_LEFT,
                    "left-" + hiderCount);
        }
    }

}
