package net.knarcraft.blockhunt.util;

import me.libraryaddict.disguise.DisguiseAPI;
import net.knarcraft.blockhunt.BlockHunt;
import net.knarcraft.blockhunt.MemoryStorage;
import net.knarcraft.blockhunt.PlayerArenaData;
import net.knarcraft.blockhunt.arena.Arena;
import net.knarcraft.blockhunt.arena.ArenaLocations;
import net.knarcraft.blockhunt.arena.ArenaState;
import net.knarcraft.blockhunt.config.ConfigKey;
import net.knarcraft.blockhunt.config.MessageKey;
import net.knarcraft.blockhunt.config.Permission;
import net.knarcraft.blockhunt.event.EndArenaEvent;
import net.knarcraft.blockhunt.event.JoinArenaEvent;
import net.knarcraft.blockhunt.event.LeaveArenaEvent;
import net.knarcraft.blockhunt.manager.PermissionsManager;
import org.bukkit.Bukkit;
import org.bukkit.GameMode;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.Sound;
import org.bukkit.World;
import org.bukkit.attribute.Attribute;
import org.bukkit.attribute.AttributeInstance;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;
import org.bukkit.inventory.meta.ItemMeta;
import org.bukkit.potion.PotionEffect;

import java.util.ArrayList;
import java.util.List;
import java.util.logging.Level;

/**
 * A helper class for dealing with arenas
 */
public final class ArenaHelper {

    private ArenaHelper() {

    }

    /**
     * Loads all arenas and creates scoreboards
     */
    public static void loadArenas() {
        MemoryStorage.arenaMap.clear();
        for (String arenaName : MemoryStorage.arenas.getFileConfiguration().getKeys(false)) {
            MemoryStorage.arenaMap.put(arenaName.toLowerCase(),
                    (Arena) MemoryStorage.arenas.getFileConfiguration().get(arenaName));
        }

        for (Arena arena : MemoryStorage.arenaMap.values()) {
            ScoreboardHelper.createScoreboard(arena);
        }
    }

    /**
     * Sends a message to all players in an arena
     *
     * @param arena     <p>The arena to announce to</p>
     * @param message   <p>The message to display</p>
     * @param variables <p>The variables to replace in the message</p>
     */
    public static void sendMessage(Arena arena, String message, String... variables) {
        MessageHelper.broadcastMessage(arena.playersInArena, message, variables);
    }

    /**
     * Sends a message to all players in an arena
     *
     * @param arena      <p>The arena to announce to</p>
     * @param messageKey <p>The message key defining the string to display</p>
     * @param variables  <p>The variables to replace in the message</p>
     */
    public static void sendMessage(Arena arena, MessageKey messageKey, String... variables) {
        MessageHelper.broadcastMessage(arena.playersInArena, messageKey, variables);
    }

    public static void playerJoinArena(Player player, String arenaName) {
        if (MemoryStorage.playerArena.get(player) != null) {
            MessageHelper.sendMessage(player, MessageKey.ERROR_JOIN_ALREADY_JOINED);
            return;
        }

        Arena arena = MemoryStorage.arenaMap.get(arenaName.toLowerCase());
        if (arena != null) {
            // TODO: Use a disabled arena state instead of checking this condition every time
            if (arena.disguiseBlocks.isEmpty()) {
                MessageHelper.sendMessage(player, MessageKey.ERROR_JOIN_NO_BLOCKS_SET);
            } else {
                joinLobby(arena, player);
            }
        } else {
            MessageHelper.sendMessage(player, MessageKey.ERROR_NO_ARENA, "name-" + arenaName);
        }

        SignsHelper.updateSigns();
    }

    public static void playerLeaveArena(Player player, boolean message, boolean cleanup) {
        Arena arena = MemoryStorage.playerArena.get(player);
        if (arena == null) {
            if (message) {
                MessageHelper.sendMessage(player, MessageKey.ERROR_LEAVE_NOT_IN_ARENA);
            }
            return;
        }

        BlockHunt.plugin.getLogger().log(Level.INFO, player.getName() + " has left " + arena.arenaName);
        LeaveArenaEvent event = new LeaveArenaEvent(player, arena);
        Bukkit.getPluginManager().callEvent(event);

        if (cleanup) {
            cleanup(arena, player);
        }

        PlayerArenaData playerArenaData = new PlayerArenaData(null, null, null,
                null, null, null, null, null, null,
                null, false);

        if (MemoryStorage.playerData.get(player) != null) {
            playerArenaData = MemoryStorage.playerData.get(player);
        } else {
            BlockHunt.plugin.getLogger().log(Level.SEVERE, "Unable to find player data for " + player.getName() +
                    ". The player's state will be broken!");
        }

        player.getInventory().clear();
        player.getInventory().setContents(playerArenaData.playerInventory());
        player.getInventory().setArmorContents(playerArenaData.playerArmor());
        player.setExp(playerArenaData.playerEXP());
        player.setLevel(playerArenaData.playerEXPLevels());
        AttributeInstance healthAttribute = player.getAttribute(Attribute.GENERIC_MAX_HEALTH);
        if (healthAttribute != null) {
            healthAttribute.setBaseValue(playerArenaData.playerMaxHealth());
        }
        player.setHealth(playerArenaData.playerHealth());
        player.setFoodLevel(playerArenaData.playerFood());
        player.addPotionEffects(playerArenaData.playerPotionEffects());
        PlayerHelper.teleport(player, arena.arenaLocations.getExitWarp());
        player.setGameMode(playerArenaData.playerGameMode());
        player.setAllowFlight(playerArenaData.playerFlying());
        player.setFlying(playerArenaData.playerFlying());
        player.setWalkSpeed(0.2F);

        MemoryStorage.playerData.remove(player);
        MemoryStorage.chosenBlock.remove(player);

        VisibilityHelper.makePlayerUnhidden(player);

        ScoreboardHelper.removeScoreboard(player);

        MessageHelper.sendMessage(player, MessageKey.NORMAL_LEAVE_YOU_LEFT);
        if (message) {
            ArenaHelper.sendMessage(arena, MessageKey.NORMAL_LEAVE_LEFT_ARENA, "%playerName%-" +
                    player.getName(), "1-" + arena.playersInArena.size(), "2-" + arena.maxPlayers);
        }

        SignsHelper.updateSigns();
    }

    private static void cleanup(Arena arena, Player player) {
        arena.playersInArena.remove(player);
        MemoryStorage.playerArena.remove(player);
        arena.seekers.remove(player);
        arena.hiders.remove(player);
        MemoryStorage.playerArena.remove(player);

        if (arena.playersInArena.size() < arena.minPlayers && arena.gameState.equals(ArenaState.STARTING)) {
            arena.gameState = ArenaState.WAITING;
            arena.timer = 0;

            sendMessage(arena, MessageKey.WARNING_LOBBY_NEED_AT_LEAST, "1-" + arena.minPlayers);
        }

        if (arena.playersInArena.size() <= 1 && arena.gameState == ArenaState.IN_GAME) {
            if (arena.seekers.size() >= arena.playersInArena.size()) {
                ArenaHelper.seekersWin(arena);
            } else {
                ArenaHelper.hidersWin(arena);
            }
        }

        if (arena.seekers.size() >= arena.playersInArena.size()) {
            ArenaHelper.seekersWin(arena);
        }

        if (arena.seekers.size() == 0 && arena.gameState == ArenaState.IN_GAME) {
            Player seeker = arena.playersInArena.get(MemoryStorage.random.nextInt(arena.playersInArena.size()));
            ArenaHelper.sendMessage(arena, MessageKey.WARNING_IN_GAME_NEW_SEEKER_CHOSEN, "seeker-" + seeker.getName());
            ArenaHelper.sendMessage(arena, MessageKey.NORMAL_IN_GAME_SEEKER_CHOSEN, "seeker-" + seeker.getName());
            DisguiseAPI.undisguiseToAll(seeker);
            for (Player pl : Bukkit.getOnlinePlayers()) {
                pl.showPlayer(BlockHunt.plugin, seeker);
            }
            seeker.getInventory().clear();
            arena.seekers.add(seeker);
            PlayerHelper.teleport(seeker, arena.arenaLocations.getSeekersWarp());
            MemoryStorage.seekerTime.put(seeker, arena.waitingTimeSeeker);
            seeker.setWalkSpeed(0.3F);

            // Fix for client not showing players after they join
            VisibilityHelper.makePlayerVisible(arena, player);
        }
    }

    /**
     * Grants rewards to the seekers, and stops the arena
     *
     * @param arena <p>The arena in which the seekers won</p>
     */
    public static void seekersWin(Arena arena) {
        String cause = "[BlockHunt] Seekers have won " + arena.arenaName;
        rewardWinnersAndStop(arena, cause, arena.seekers, arena.seekersWinCommands, arena.seekersTokenWin,
                MessageKey.NORMAL_WIN_SEEKERS);
    }

    /**
     * Grant rewards to the hiders, and stops the arena
     *
     * @param arena <p>The arena in which the hiders won</p>
     */
    public static void hidersWin(Arena arena) {
        String cause = "[BlockHunt] Hiders have won " + arena.arenaName;
        rewardWinnersAndStop(arena, cause, arena.hiders, arena.hidersWinCommands, arena.hidersTokenWin,
                MessageKey.NORMAL_WIN_HIDERS);
    }

    /**
     * Rewards winners of an arena and stops the arena
     *
     * @param arena       <p>The arena to grant rewards for</p>
     * @param cause       <p>The cause of the arena stopping</p>
     * @param winnerTeam  <p>The players on the team that won</p>
     * @param winCommands <p>The commands to be executed for the winning team</p>
     * @param winTokens   <p>The amount of tokens to grant to the winners</p>
     * @param winMessage  <p>The message to display to the players in the arena regarding which team won</p>
     */
    private static void rewardWinnersAndStop(Arena arena, String cause, List<Player> winnerTeam,
                                             List<String> winCommands, int winTokens, MessageKey winMessage) {

        List<Player> winners = new ArrayList<>();
        List<Player> losers = new ArrayList<>();

        for (Player player : arena.playersInArena) {
            if (winnerTeam.contains(player)) {
                winners.add(player);
                grantRewards(player, winCommands, winTokens);
            } else {
                losers.add(player);
            }
        }

        // Stop the arena
        stopArena(arena, winners, losers, cause, winMessage);
    }

    /**
     * Stops an arena
     *
     * @param arena   <p>The arena to stop</p>
     * @param winners <p>The winners of the arena</p>
     * @param losers  <p>The losers of the arena</p>
     * @param cause   <p>The cause of why the arena was stopped</p>
     * @param message <p>The message to display to players</p>
     */
    private static void stopArena(Arena arena, List<Player> winners, List<Player> losers, String cause,
                                  MessageKey message) {
        EndArenaEvent event = new EndArenaEvent(winners, losers, arena);
        Bukkit.getServer().getPluginManager().callEvent(event);
        stopArena(arena, cause, message);
    }

    /**
     * Grant arena win rewards to a player
     *
     * @param player      <p>The player to reward</p>
     * @param winCommands <p>The commands to execute upon winning</p>
     * @param tokens      <p>The amount of tokens to grant</p>
     */
    private static void grantRewards(Player player, List<String> winCommands, int tokens) {

        // Execute any hiders win commands
        if (winCommands != null) {
            for (String command : winCommands) {
                Bukkit.dispatchCommand(Bukkit.getConsoleSender(), command.replaceAll("%player%",
                        player.getDisplayName()));
            }
        }

        // Give the player their tokens
        int playerTokens = TokensHelper.getTokens(player);
        TokensHelper.setTokens(player, playerTokens + tokens);

        MessageHelper.sendMessage(player, MessageKey.NORMAL_ADDED_TOKEN, "amount-" + tokens);
    }

    /**
     * Stops an arena
     *
     * @param arena   <p>The arena to stop</p>
     * @param cause   <p>A description of why the arena was stopped</p>
     * @param message <p>The message to display to affected players</p>
     */
    public static void stopArena(Arena arena, String cause, MessageKey message) {
        BlockHunt.plugin.getLogger().log(Level.INFO, cause);
        ArenaHelper.sendMessage(arena, message);
        arena.seekers.clear();
        arena.hiders.clear();

        for (Player player : arena.playersInArena) {
            playerLeaveArena(player, false, false);
            MemoryStorage.playerArena.remove(player);
            player.playSound(player.getLocation(), Sound.ENTITY_PLAYER_LEVELUP, 1, 1);
        }

        arena.gameState = ArenaState.WAITING;
        arena.timer = 0;
        arena.playersInArena.clear();
    }

    /**
     * Checks whether any arena has any player
     *
     * @return <p>True if all arenas are empty</p>
     */
    public static boolean noPlayersInArenas() {
        // Check if there are any players in any arena (quick way to early exit for event handlers)
        for (Arena arena : MemoryStorage.arenaMap.values()) {
            if (arena.playersInArena.size() > 0) {
                return false;
            }
        }
        return true;
    }

    /**
     * Makes the given player join the given arena's lobby
     *
     * @param arena  <p>The arena the player wants to join</p>
     * @param player <p>The player that's joining the arena</p>
     */
    private static void joinLobby(Arena arena, Player player) {
        // Deny the player if their inventory needs to be empty, and it's not
        if ((Boolean) MemoryStorage.config.get(ConfigKey.REQUIRE_INVENTORY_CLEAR_ON_JOIN) && !isInventoryEmpty(player)) {
            MessageHelper.sendMessage(player, MessageKey.ERROR_JOIN_INVENTORY_NOT_EMPTY);
            return;
        }

        // Check that the arena has been properly set up before allowing the player to play
        // TODO: It would be better to have a DISABLED state for arenas until they are properly set up.
        ArenaLocations locations = arena.arenaLocations;
        if (!areWarpsSet(locations, player.getWorld())) {
            MessageHelper.sendMessage(player, MessageKey.ERROR_JOIN_WARPS_NOT_SET);
            return;
        }

        // Make sure the game isn't already running
        if (arena.gameState == ArenaState.IN_GAME) {
            MessageHelper.sendMessage(player, MessageKey.ERROR_JOIN_ARENA_IN_GAME);
            return;
        }

        // Check if the arena is already full
        if (arena.playersInArena.size() >= arena.maxPlayers &&
                !PermissionsManager.hasPermission(player, Permission.JOIN_FULL, false)) {
            MessageHelper.sendMessage(player, MessageKey.ERROR_JOIN_FULL);
            return;
        }

        // Teleport the player to the arena
        boolean canWarp = PlayerHelper.teleport(player, locations.getLobbyWarp());
        if (!canWarp) {
            MessageHelper.sendMessage(player, MessageKey.ERROR_TELEPORT_FAILED);
            return;
        }

        BlockHunt.plugin.getLogger().log(Level.INFO, player.getName() + " has joined " + arena.arenaName);
        arena.playersInArena.add(player);
        MemoryStorage.playerArena.put(player, arena);
        // Call the join event for any add-ons
        Bukkit.getPluginManager().callEvent(new JoinArenaEvent(player, arena));

        AttributeInstance healthAttribute = player.getAttribute(Attribute.GENERIC_MAX_HEALTH);

        // Store the player's data for restoration later
        PlayerArenaData playerArenaData = new PlayerArenaData(player.getLocation(), player.getGameMode(),
                player.getInventory().getContents(), player.getInventory().getArmorContents(), player.getExp(),
                player.getLevel(), healthAttribute != null ? healthAttribute.getValue() : 20, player.getHealth(),
                player.getFoodLevel(), player.getActivePotionEffects(), player.getAllowFlight());
        MemoryStorage.playerData.put(player, playerArenaData);

        // Set the player's state to the desired state
        setArenaState(arena, player, healthAttribute);

        // Fix for client not showing players after they join
        VisibilityHelper.makePlayerVisible(arena, player);

        // Give the player access to the block chooser
        giveBlockChooser(player);

        // Give the player access to the hider/seeker chooser
        giveHiderSeekerChooser(player);

        // Remove any existing disguise
        DisguiseAPI.undisguiseToAll(player);

        // Announce that the player joined
        ArenaHelper.sendMessage(arena, MessageKey.NORMAL_JOIN_JOINED_ARENA, "%playerName%-" + player.getName(),
                "1-" + arena.playersInArena.size(), "2-" + arena.maxPlayers);
        if (arena.playersInArena.size() < arena.minPlayers) {
            // Announce that additional players are required for the arena to start
            sendMessage(arena, MessageKey.WARNING_LOBBY_NEED_AT_LEAST, "1-" + arena.minPlayers);
        }
    }

    /**
     * Gives a hider/seeker chooser to the given player if necessary
     *
     * @param player <p>The player to give the hider/seeker chooser to</p>
     */
    private static void giveHiderSeekerChooser(Player player) {
        // If not enabled, return
        if (!((Boolean) MemoryStorage.config.get(ConfigKey.SHOP_BLOCK_HUNT_PASS_V_2_ENABLED))) {
            return;
        }

        // If the player is out of passes, don't give them the choice between hider and seeker
        if (MemoryStorage.shop.getFileConfiguration().getInt(player.getName() + ".blockhuntpass") <= 0) {
            return;
        }

        Material chooserMaterial = Material.matchMaterial((String) MemoryStorage.config.get(
                ConfigKey.SHOP_BLOCK_HUNT_PASS_V_2_ID_NAME));
        if (chooserMaterial == null) {
            BlockHunt.plugin.getLogger().log(Level.SEVERE, "Cannot give the block chooser because an invalid" +
                    " material has been set.");
            return;
        }
        ItemStack shopBlockHuntPass = new ItemStack(chooserMaterial, 1);
        ItemMeta shopBlockHuntPassItemMeta = shopBlockHuntPass.getItemMeta();
        if (shopBlockHuntPassItemMeta == null) {
            return;
        }
        shopBlockHuntPassItemMeta.setDisplayName(MessageHelper.replaceAll((String) MemoryStorage.config.get(
                ConfigKey.SHOP_BLOCK_HUNT_PASS_V_2_NAME)));
        List<String> loreStrings = MemoryStorage.config.getFileConfiguration().getStringList(
                ConfigKey.SHOP_BLOCK_HUNT_PASS_V_2_DESCRIPTION.getPath());
        List<String> loreStrings2 = new ArrayList<>();
        for (String lore : loreStrings) {
            loreStrings2.add(MessageHelper.replaceAll(lore));
        }

        shopBlockHuntPassItemMeta.setLore(loreStrings2);
        shopBlockHuntPass.setItemMeta(shopBlockHuntPassItemMeta);
        shopBlockHuntPass.setAmount(MemoryStorage.shop.getFileConfiguration().getInt(
                player.getName() + ".blockhuntpass"));

        player.getInventory().addItem(shopBlockHuntPass);
    }

    /**
     * Gives a block chooser to the given player if necessary
     *
     * @param player <p>The player to give the block chooser to</p>
     */
    private static void giveBlockChooser(Player player) {
        // If missing the necessary permission or if it's disabled, abort
        if (!((Boolean) MemoryStorage.config.get(ConfigKey.SHOP_BLOCK_CHOOSER_V_1_ENABLED)) ||
                (MemoryStorage.shop.getFileConfiguration().get(player.getName() + ".blockchooser") == null &&
                        !PermissionsManager.hasPermission(player, Permission.SHOP_BLOCK_CHOOSER, false))) {
            return;
        }

        Material chooserMaterial = Material.matchMaterial((String) MemoryStorage.config.get(
                ConfigKey.SHOP_BLOCK_CHOOSER_V_1_ID_NAME));
        if (chooserMaterial == null) {
            BlockHunt.plugin.getLogger().log(Level.SEVERE, "Cannot give the block chooser because an invalid" +
                    " material has been set.");
            return;
        }

        ItemStack shopBlockChooser = new ItemStack(chooserMaterial, 1);
        ItemMeta shopBlockChooserItemMeta = shopBlockChooser.getItemMeta();
        if (shopBlockChooserItemMeta == null) {
            return;
        }
        shopBlockChooserItemMeta.setDisplayName(MessageHelper.replaceAll((String) MemoryStorage.config.get(
                ConfigKey.SHOP_BLOCK_CHOOSER_V_1_NAME)));
        List<String> loreStrings = MemoryStorage.config.getFileConfiguration().getStringList(
                ConfigKey.SHOP_BLOCK_CHOOSER_V_1_DESCRIPTION.getPath());
        List<String> loreStrings2 = new ArrayList<>();
        for (String loreLine : loreStrings) {
            loreStrings2.add(MessageHelper.replaceAll(loreLine));
        }
        shopBlockChooserItemMeta.setLore(loreStrings2);
        shopBlockChooser.setItemMeta(shopBlockChooserItemMeta);

        player.getInventory().addItem(shopBlockChooser);
    }

    /**
     * Sets the player's state to the arena's required state
     *
     * <p>Sets food and health to defaults, and clears the inventory among other things.</p>
     *
     * @param arena           <p>The arena the player joined</p>
     * @param player          <p>The player to update the state for</p>
     * @param healthAttribute <p>The player's health attribute</p>
     */
    private static void setArenaState(Arena arena, Player player, AttributeInstance healthAttribute) {
        player.setGameMode(GameMode.SURVIVAL);
        for (PotionEffect potionEffect : player.getActivePotionEffects()) {
            player.removePotionEffect(potionEffect.getType());
        }
        player.setFoodLevel(20);
        player.setHealth(20);

        if (healthAttribute != null) {
            healthAttribute.setBaseValue(20);
        }
        player.setLevel(arena.timer);
        player.setExp(0);
        player.getInventory().clear();
        player.getInventory().setHelmet(new ItemStack(Material.AIR));
        player.getInventory().setChestplate(new ItemStack(Material.AIR));
        player.getInventory().setLeggings(new ItemStack(Material.AIR));
        player.getInventory().setBoots(new ItemStack(Material.AIR));
        player.setFlying(false);
        player.setAllowFlight(false);
        player.setWalkSpeed(0.2F);
    }

    /**
     * Checks whether the given player's inventory is completely empty
     *
     * @param player <p>The player to check</p>
     * @return <p>True if the player's inventory is completely empty</p>
     */
    private static boolean isInventoryEmpty(Player player) {
        for (ItemStack inventoryItem : player.getInventory()) {
            if (inventoryItem != null && inventoryItem.getType() != Material.AIR) {
                return false;
            }
        }
        for (ItemStack inventoryItem : player.getInventory().getArmorContents()) {
            if (inventoryItem != null && inventoryItem.getType() != Material.AIR) {
                return false;
            }
        }
        return true;
    }

    /**
     * Checks whether all warps are set in the given ArenaLocations object
     *
     * @param locations <p>The arena locations to check</p>
     * @param world     <p>The world the arena belongs to</p>
     * @return <p>True if all warps have been properly set</p>
     */
    private static boolean areWarpsSet(ArenaLocations locations, World world) {
        Location zero = new Location(world, 0, 0, 0, 0, 0);
        return locations.getLobbyWarp() != null && locations.getHidersWarp() != null &&
                locations.getSeekersWarp() != null && locations.getExitWarp() != null &&
                !locations.getLobbyWarp().equals(zero) && !locations.getHidersWarp().equals(zero) &&
                !locations.getSeekersWarp().equals(zero) && !locations.getExitWarp().equals(zero);
    }

}
