package net.knarcraft.blockhunt.util;

import me.libraryaddict.disguise.DisguiseAPI;
import me.libraryaddict.disguise.disguisetypes.DisguiseType;
import me.libraryaddict.disguise.disguisetypes.MiscDisguise;
import net.knarcraft.blockhunt.BlockHunt;
import net.knarcraft.blockhunt.MemoryStorage;
import net.knarcraft.blockhunt.arena.Arena;
import net.knarcraft.blockhunt.config.MessageKey;
import org.bukkit.Bukkit;
import org.bukkit.ChatColor;
import org.bukkit.Location;
import org.bukkit.Material;
import org.bukkit.Sound;
import org.bukkit.block.Block;
import org.bukkit.block.data.BlockData;
import org.bukkit.enchantments.Enchantment;
import org.bukkit.entity.Player;
import org.bukkit.inventory.ItemStack;

import java.util.Objects;
import java.util.logging.Level;

import static net.knarcraft.blockhunt.util.StringHelper.uppercaseAllFirst;

/**
 * A helper class for editing player visibility
 */
public final class VisibilityHelper {

    private VisibilityHelper() {

    }

    /**
     * Makes a player visible to players in an arena and the other way around
     *
     * @param arena  <p>The arena the player was in</p>
     * @param player <p>The player to make visible</p>
     */
    public static void makePlayerVisible(Arena arena, Player player) {
        for (Player otherPlayer : arena.playersInArena) {
            if (otherPlayer.canSee(player)) {
                // Make player visible to others
                otherPlayer.showPlayer(BlockHunt.plugin, player);
            }
            if (player.canSee(otherPlayer)) {
                // Make other players visible to player
                player.showPlayer(BlockHunt.plugin, otherPlayer);
            }
        }
    }

    /**
     * Makes a hidden player in an arena no longer appear as hidden
     *
     * <p>This should be used when a player leaves an arena or the server to make sure no remnants of the hidden block
     * remains</p>
     *
     * @param player <p>The player to un-hide</p>
     */
    public static void makePlayerUnhidden(Player player) {
        // Un-disguise the player
        DisguiseAPI.undisguiseToAll(player);

        // Get the block data for the player
        BlockData hiddenBlockData = null;
        if (MemoryStorage.hiddenLocationMaterial.get(player) != null) {
            hiddenBlockData = Bukkit.createBlockData(MemoryStorage.hiddenLocationMaterial.get(player));
        }

        for (Player onlinePlayer : Bukkit.getOnlinePlayers()) {
            // Make the player visible to other players
            onlinePlayer.showPlayer(BlockHunt.plugin, player);

            // No longer display the player's solid block
            if (hiddenBlockData != null) {
                Block playerBlock = MemoryStorage.hiddenLocation.get(player).getBlock();
                onlinePlayer.sendBlockChange(playerBlock.getLocation(), hiddenBlockData);
            }
        }
    }

    /**
     * Makes the player into a no longer solid block
     *
     * @param player <p>The player to make unsolid</p>
     */
    public static void makePlayerUnSolid(Player player) {
        ItemStack block = player.getInventory().getItem(8);
        if (block == null) {
            // This shouldn't be possible, but just in case
            player.sendMessage(ChatColor.RED + "Unable to hide you because your inventory block is missing!");
            BlockHunt.plugin.getLogger().log(Level.SEVERE, player.getName() + " could not be hidden because " +
                    "their inventory block was missing!");
            return;
        }

        Block playerBlock = player.getLocation().getBlock();
        if (MemoryStorage.hiddenLocation.get(player) != null) {
            playerBlock = MemoryStorage.hiddenLocation.get(player).getBlock();
        }

        block.setAmount(5);
        for (Player otherPlayer : Bukkit.getOnlinePlayers()) {
            if (otherPlayer.equals(player)) {
                continue;
            }

            // Restore the block the player replaced
            Material material = MemoryStorage.hiddenLocationMaterial.get(player);
            otherPlayer.sendBlockChange(playerBlock.getLocation(),
                    Bukkit.createBlockData(Objects.requireNonNullElse(material, Material.AIR)));
            MemoryStorage.hiddenLocationMaterial.remove(player);
        }

        player.playSound(player.getLocation(), Sound.ENTITY_BAT_HURT, 1, 1);
        block.removeEnchantment(Enchantment.DURABILITY);

        for (Player playerShow : Bukkit.getOnlinePlayers()) {
            playerShow.showPlayer(BlockHunt.plugin, player);
        }

        MiscDisguise disguise = new MiscDisguise(DisguiseType.FALLING_BLOCK, block.getType());
        DisguiseAPI.disguiseToAll(player, disguise);

        MessageHelper.sendMessage(player, MessageKey.NORMAL_IN_GAME_NO_MORE_SOLID);
    }

    /**
     * Tries to make the given player solid
     *
     * @param player     <p>The player to make solid</p>
     * @param solidBlock <p>The block to solidify the player as</p>
     */
    public static void makePlayerSolid(Player player, ItemStack solidBlock) {
        Block playerBlock = player.getLocation().getBlock();
        BlockData playerBlockData = playerBlock.getBlockData();
        if (playerBlockData.getMaterial().isAir() || !playerBlockData.getMaterial().isSolid()) {
            makePlayerSolid(player, playerBlock, playerBlockData, solidBlock, player.getLocation());
        } else {
            MessageHelper.sendMessage(player, MessageKey.WARNING_IN_GAME_NO_SOLID_PLACE);
        }
    }

    /**
     * Makes a player into a perceived solid block
     *
     * @param player          <p>The player to make solid</p>
     * @param playerBlock     <p>The block the player is currently standing at</p>
     * @param playerBlockData <p>The data of the block the player is currently standing at</p>
     * @param inventoryBlock  <p>The inventory block which corresponds to the player's hide material</p>
     * @param playerLocation  <p>The current location of the player</p>
     */
    private static void makePlayerSolid(Player player, Block playerBlock, BlockData playerBlockData, ItemStack inventoryBlock,
                                        Location playerLocation) {
        // Store the material the player replaced
        MemoryStorage.hiddenLocationMaterial.put(player, playerBlockData.getMaterial());

        if (DisguiseAPI.isDisguised(player)) {
            // Hide the player, and make all other players see a full block in the player's location
            DisguiseAPI.undisguiseToAll(player);
            for (Player onlinePlayer : Bukkit.getOnlinePlayers()) {
                if (!onlinePlayer.equals(player)) {
                    onlinePlayer.hidePlayer(BlockHunt.plugin, player);
                    onlinePlayer.sendBlockChange(playerBlock.getLocation(),
                            inventoryBlock.getType().createBlockData());
                }
            }

            inventoryBlock.addUnsafeEnchantment(Enchantment.DURABILITY, 10);
            player.playSound(playerLocation, Sound.ENTITY_EXPERIENCE_ORB_PICKUP, 1, 1);
            // Store the location the player is hidden at
            MemoryStorage.hiddenLocation.put(player, playerLocation);

            // Announce to the player that they are now hidden
            String blockName = inventoryBlock.getType().name();
            blockName = uppercaseAllFirst(blockName.replace("_", " "));
            MessageHelper.sendMessage(player, MessageKey.NORMAL_IN_GAME_NOW_SOLID, "block-"
                    + blockName);
        }

        for (Player otherPlayer : Bukkit.getOnlinePlayers()) {
            if (otherPlayer.equals(player)) {
                continue;
            }

            // Hide the player and update the perceived
            otherPlayer.hidePlayer(BlockHunt.plugin, player);
            otherPlayer.sendBlockChange(playerBlock.getLocation(), inventoryBlock.getType().createBlockData());
        }
    }

}
